#!/usr/bin/perl
# $Id$

# name: spamc-learn-wrapper.pl
# mod:  2020-06-12

use strict;

# cpanp -i MIME::Tools MIME::Parser Net::IP Getopt::Long Pod::Usage File::Temp
use File::Temp;
use Getopt::Long;
use MIME::Parser;
use MIME::Tools;
use Net::IP;
use POSIX;

##########################
## postfix integration

# SA-LEARN: Bayes Spam/Ham Training


# @ /usr/local/etc/postfix/master.cf
## SA-LEARN: Bayes Spam/Ham Training
#	sa-spam  unix  -  n  n  -  -  pipe
#	# user=spamd:spamd argv=/usr/local/etc/spamassassin/scripts/sa-learn-wrapper.pl spam ${sender}
#	  user=spamd:spamd
#	  argv=/usr/local/etc/spamassassin/scripts/spamc-learn-wrapper.pl -L spam
#	  flags=DX
#	sa-ham  unix  -  n  n  -  -  pipe
#	# user=spamd:spamd argv=/usr/local/etc/spamassassin/scripts/sa-learn-wrapper.pl ham  ${sender}
#	  user=spamd:spamd
#	  argv=/usr/local/etc/spamassassin/scripts/spamc-learn-wrapper.pl -L ham
#	  flags=DX


# !! NOTE !!
# 	after script change, exec
# systemctl restart postfix

##########################
## config params

# 0:quiet -> 3:debug
my $DEBUG = 2;

#my $SPAMC    = '/usr/bin/spamc';
#my $SA_LEARN = '/usr/bin/sa-learn';
my $SPAMC    = '/usr/bin/perl -I/usr/local/postfix-pgnd/spamassassin/lib/perl5  /usr/local/postfix-pgnd/spamassassin/bin/spamc';
my $SA_LEARN = '/usr/bin/perl -I/usr/local/postfix-pgnd/spamassassin/lib/perl5  /usr/local/postfix-pgnd/spamassassin/bin/sa-learn';

my $UNPACK_DIR = '/var/lib/spamassassin/unpack';

# 1: log to $LOGFILE
# 0: log to STDOUT
my $LOG_TO_FILE = 1;
my $LOGFILE = '/var/log/spamd/sa-learn.log';

# 1: delete msgs/dirs from $UNPACK_DIR
# 0: leave unpacked msgs/dirs in $UNPACK_DIR
my $CLEANUP = 0;

# postconf -n mynetworks
#  mynetworks = $var_MX1_EDGE_IPv4/32 $var_MX2_EDGE_IPv4/32 $var_MX1_EDGE_IPv4_INT/32 $var_MX1_CORE_IPv4_INT/32 127.0.0.0/8
# plus, '10.0.0.1', which is the spoofed IP from mx1-core postfix instance
my @MYNETWORKS = qw(
  209.123.234.53/32
  209.123.234.54/32
  172.30.7.100/32
  172.30.11.50/32
  127.0.0.0/8
  10.0.0.1/32
);


##########################
## vars
my @IPinNetworks;
my $msg_parser = new MIME::Parser;
my $log_string;
my $tmp_log = File::Temp::tmpnam(); #"$UNPACK_DIR/__log.tmp";
my $counter = 1;
my ($spamham, $client);


##########################
## opts
Getopt::Long::Configure ('no_ignore_case');
if ( @ARGV > 0) {
	GetOptions(
		'debug=i'    => \$DEBUG,
		'd+'         => \$DEBUG,
		'L=s'        => \$spamham
	)
	or pod2usage(1);
}


##########################
## check opts
if ( $spamham =~ m/^(?:spam|ham|forget)$/i ) {
	$spamham = "-L $spamham";
	$client = $SPAMC;
}

##########################
## logs
# open the logfile for writing and share (1) it.
if ($LOG_TO_FILE) {
	open (DELOG, ">>$LOGFILE") or die $!;
	flock (DELOG, 1);
}

# Debug level 3
if ($DEBUG >= 3) {
	MIME::Tools->debugging(1);
	open(STDERR, ">>$LOGFILE");
}


##########################
## subs
sub failure {
	# Debug Level 1
	$log_string = "--f " . $_[0];
	logging("--f " . $_[0], 1);
	if ($LOG_TO_FILE) {
		close(DELOG);
	}
	die $_[0];
}

sub logging {
	my($message, $log_level) = @_;
	# Jan  9 21:03:43
	my $now = strftime "%b %e %H:%M:%S", localtime;
	chomp($message);
	$message =~ s/\n/\n$now /g;
	if ($LOG_TO_FILE == 1) {
		if ($log_level <= $DEBUG) {
			print DELOG $now ." " .$message ."\n";
		}
	} else {
		if ($log_level <= $DEBUG) {
			print $now ." " .$message ."\n";
		}
	}
}

sub open_tmp_log {
	open (TMP, "+>>$tmp_log") || failure("failed to open $tmp_log!");
	flock (TMP, 1);
}

sub read_close_tmp_log {
	my $log_level = shift;
	my $string = shift;
	seek(TMP, 0, 0);	# go to the beginning of $tmp_log file
	my @dump = <TMP>;
	close(TMP) || failure("failed to close $tmp_log!");
	if (defined($string)) {
		map {
			if ($_ =~ $string) {
				logging(
					"$_",
					$log_level
				);
			}
		} @dump;
	} else {
		map {
			logging(
				"$_",
				$log_level
			);
		} @dump;
	}
	unlink($tmp_log);
}

sub cleanup {
	if ($CLEANUP) {
		# Debug Level 1
		logging(
			"--: cleanup now",
			1
		);
		# remove the unpacked message files
		$msg_parser->filer->purge;
		# remove the created output directory
		rmdir $msg_parser->output_dir;
	} else {
		logging(
			"--! will NOT cleanup",
			0
		);
	}

	if ($LOG_TO_FILE) {
		close(DELOG);
	}
}

sub submit {
	my $ent = shift;

	if (! $ent->parts) {
		# Debug Level 2
		logging("--: part number: " . $counter++, 2);
		logging($ent->body_as_string(), 2);
		#       $ent->print(\*DELOG); # don't print to FH
	}

	if ($ent->head->mime_type eq 'message/rfc822') {
		logging("--! msg piped to: $client", 0);
		# Debug Level 3
		logging("--: DUMP piped msg:", 3);
		#	    $ent->bodyhandle->print(\*DELOG);	# don't print to FH
		logging($ent->bodyhandle->as_string, 3); # or ->as_lines
		logging("--: end msg DUMP", 3);

		if ( $spamham != m/^-L/ ) {
			if ($DEBUG >= 3) { # Debug Level 3
				$spamham = "-D $spamham";
			} elsif ($DEBUG >= 1) { # Debug Level 1
				$spamham = "-D learn $spamham";
			} # else: Debug Level 0
		}

		open(SALEARN, "|$client $spamham >$tmp_log 2>&1") ||
		     failure("error, pipe to $client: $!");
			$ent->bodyhandle->print(\*SALEARN) &&
		     logging("--! msg pipe OK.", 0);
		close(SALEARN);

		# workaround; $client only logs to a file
		open_tmp_log();
		read_close_tmp_log(1);

	}
	#  elsif ($ent->head->mime_type eq 'text/html')
	#  {
	#     print "body of type message:\n";
	#     $msg_submit->head->print(\*STDOUT);
	#     $msg_submit->bodyhandle->print(\*STDOUT);
	#  }

	my @parts = $ent->parts;
	if (@parts) {
		map {
			submit($_);
		} @parts;
	} elsif ($counter == 1) {
		logging(
			"--! There aren't any parts "
			."-> Nothing to submit\n--! quit",
			0
		);
	}
}

##########################
## ip list/check
logging(
	"##### Starting sal-wrapper ... #####",
	0
);

#mkdir -p $UNPACK_DIR or failure($UNPACK_DIR ." create FAILed", 0);
mkdir -p $UNPACK_DIR;
# https://superuser.com/questions/624449/chown-function-in-perl
my $_uid = getpwnam("spamd");
my $_gid = getgrnam("spamd");
chown $_uid, $_gid, $UNPACK_DIR;

map {
	my $ip = new Net::IP($_) || failure("error, IP object");
	@IPinNetworks = (@IPinNetworks, $ip);
} @MYNETWORKS;


# Debug level 2
logging("--: list LOCAL IP addr:", 2);
map {
	logging(
		$_->print(),
		2
	);
} @IPinNetworks;

# don't decode the content of a message/rfc822 document
$msg_parser->extract_nested_messages(0);
$msg_parser->output_under($UNPACK_DIR);

my $msg_submit = $msg_parser->parse(\*STDIN) or
             failure("parse error, STDIN");

# Debug level 2
logging(
	"--: DUMP msg skeleton:",
	2
);
# work around; dump_skeleton can only dump to a FH
if ($DEBUG >=2) {
	open_tmp_log();
	$msg_submit->dump_skeleton(\*TMP);
	read_close_tmp_log(2);
	logging(
		"--: end DUMP msg skeleton",
		2
	);
}

#Debug level 3
logging(
	"--: found the following \'Received:\' lines:",
	3
);

# parse sender IP addr @ Received: header
for (my $i=0; $i<$msg_submit->head->count('Received'); $i++) {
	my $LocalMail=0;
	my $rcvd_line = $msg_submit->head->get('Received',$i);
	# Debug level 3
	logging(
		"$rcvd_line",
		3
	);

	if ($rcvd_line =~ /^from [a-zA-Z0-9_.-]* \([a-zA-Z0-9_.-]* ?\[((?:[0-9]{1,3}\.){3}[0-9]{1,3})\](?: verified)?\)/ ||
	  $rcvd_line =~ /^from \[((?:[0-9]{1,3}\.){3}[0-9]{1,3})\] \(.*\)/) {

		# Debug level 2
		logging(
			"--: extracted IP: $1",
			2
		);
		my $ip = new Net::IP($1) ||
		         failure("creating IP object failed");
		map {
			if ($ip->overlaps($_)==$IP_A_IN_B_OVERLAP ||
			    $ip->overlaps($_)==$IP_IDENTICAL) {

				# Debug level 1
				logging(
					"--: " .$_->print()
					." contains " .$ip->print()
					." -> sent from LOCAL net\n",
					1
				);
				$LocalMail = 1;
			 }
		} @IPinNetworks;

		if (! $LocalMail) {
			logging(
				"--! NOT sent from LOCAL net",
				0
			);
			logging(
				"--! EXITing.",
				0
			);
			cleanup;
			exit;
		}

	} else {
		logging(
			"--! no IP addr found!",
			0
		);
	}
}

if (! @IPinNetworks) {
	cleanup;
	failure("No IPs in my network!");
}


##########################
## submission
# root (0/header) already examined
# submit rest of piped message to
# spamc if MIME type == message/rfc822
submit($msg_submit);


##########################
## end
cleanup;

